#!/usr/bin/env python3
# Copyright (c) 2011 Citrix Systems, Inc. All use and distribution of this
# copyrighted material is governed by and subject to terms and conditions
# as licensed by Citrix Systems, Inc. All other rights reserved.
# Xen, XenSource and XenEnterprise are either registered trademarks or
# trademarks of Citrix Systems, Inc. in the United States and/or other
# countries.

import logging
import os
import os.path
import re
import shutil
import socket
import stat
import urllib.parse as urlparse
from io import BytesIO
import json

import xcp.accessor
import xcp.bootloader as bootloader
import xcp.cmd as cmd
import xcp.cpiofile as cpiofile
import xcp.dom0 as dom0
import xcp.pci as pci
import xcp.repository as repository
import xcp.version as version
import xcp.logger as logger
import XenAPI
import XenAPIPlugin
import xen.lowlevel.xc

min_upgrade_lvm_part_size = 38 * 2**30 #38GB

boot_files = ['install.img', 'boot/vmlinuz', 'boot/xen.gz', 'boot/isolinux/isolinux.cfg']
xs_6_2 = version.Version([1, 8, 0])
xs_6_5 = version.Version([1, 9, 0])
xs_7_0 = version.Version([2, 1, 0])
xs_8_C = version.Version([3, 2, 50])
xs_min_version = xs_7_0

network_schemes = ['http', 'https', 'ftp', 'nfs']

def sanitize_url(url):
    '''Replace username/password in a URL with asterisks for display'''

    res = urlparse.urlsplit(url)

    if res.username is not None and res.password is not None:
        return url.replace('%s:%s@' % (res.username, res.password), '***:***@', 1)
    elif res.username is not None:
        return url.replace('%s@' % res.username, '***:***@', 1)
    else:
        # Cannot have a password without a username
        assert res.password is None

    return url

def process_error_code(c):
    if c == 401:
        raise Exception('UNAUTH_URL')
    elif c == 403:
        raise Exception('FORBIDDEN')
    elif c == 404:
        raise Exception('NOT_FOUND')
    elif c == 407:
        raise Exception('PROXY_AUTH_FAILURE')
    elif c == 408:
        raise Exception('TIMEOUT')
    elif c >= 500 and c <= 511:
        raise Exception('REPO_SERVER_ERROR_' + c)
    elif c == TEST_URL_INVALID:
        raise Exception('INVALID_URL')
    elif c == TEST_VER_INVALID:
        raise Exception('INVALID_VER')
    elif c == 5:
        raise Exception('IOERROR')
    else:
        raise Exception('ERROR code: ', c)

def shell_value(line):
    return line.split('=', 1)[1].strip("'")

def without_scheme(url):
    return url.split('//', 1)[1]

def test_boot_files(accessor):
    done = TEST_REPO_GOOD
    accessor.start()
    for f in boot_files:
        try:
            logger.info("Testing "+f)
            if accessor.access(f):
                logger.info("    success")
            else:
                logger.error("    failed")
                logger.error(accessor.lastError)
                done = accessor.lastError
                break
        except Exception as e:
            logger.logException(e)
            raise Exception('UNKNOWN_ERROR')

    accessor.finish()
    return done

def get_boot_files(accessor, dest_dir):
    accessor.start()
    try:
        for f in boot_files:
            try:
                logger.info("Fetching " + f)
                accessor.lastError = 0
                inf = accessor.openAddress(f)
                outf = open(os.path.join(dest_dir, os.path.basename(f)), 'wb')
                outf.writelines(inf)
                outf.close()
                inf.close()
            except Exception as e:
                logger.logException(e)
                if accessor.lastError != 0:
                    process_error_code(accessor.lastError)
                raise Exception('IOERROR')
    finally:
        accessor.finish()

def map_label_to_partition(label):
    partition = None
    (rc, out) = cmd.runCmd(['blkid', '-l', '-t', 'LABEL="%s"' % label, '-o', 'device'],
                           with_stdout=True)
    if rc == 0 and out.startswith('/dev/'):
        partition = out.strip()
        if os.path.isfile('/sbin/udevadm'):
            args = ['/sbin/udevadm', 'info']
        else:
            args = ['udevinfo']
        (rc, out) = cmd.runCmd(args + ['-q', 'symlink', '-n', partition[5:]],
                               with_stdout=True)
        if rc == 0:
            for link in out.split():
                if link.startswith('disk/by-id') and not link.startswith('disk/by-id/edd'):
                    partition = '/dev/'+link
                    break

    return partition

def get_fs_labels():
    root_label = None
    boot_label = None
    try:
        # determine root and boot labels
        f = open('/etc/fstab')
        for line in f:
            line = line.strip()
            if line.startswith('LABEL=root-'):
                v, _ = line.split(None, 1)
                root_label = v[6:]
            if line.startswith('LABEL=BOOT-'):
                v, _ = line.split(None, 1)
                boot_label = v[6:]
        f.close()
    except Exception as e:
        logger.logException(e)
        logger.error("Failed to read fstab")

    return root_label, boot_label

def get_md_uuid(dev):
    _, out = cmd.runCmd(['mdadm', '--detail', '--export', dev],
                        with_stdout=True)
    for line in out.split("\n"):
        line = line.strip().split('=', 1)
        if line[0] == 'MD_UUID':
            return line[1]

    return None

def gen_answerfile(accessor, installer_dir, url, updates, scripts, backup_config, args):
    root_device = None
    root_partition = None
    boot_partition = None

    try:
        # determine root disk
        f = open('/etc/xensource-inventory')
        for l in f:
            line = l.strip()
            if line.startswith('PRIMARY_DISK='):
                root_device = shell_value(line)
                break
        f.close()
    except Exception as e:
        logger.logException(e)
        logger.error("Failed to read inventory")
        raise Exception('DISK_PROBE_ERROR')
    if not root_device:
        logger.error("Failed to determine root disk")
        raise Exception('DISK_PROBE_ERROR')

    if not os.path.exists(root_device):
        logger.error("Root disk %s not found" % root_device)
        raise Exception('DISK_PROBE_ERROR')

    for script in scripts:
        if not os.path.exists(script['path']):
            logger.error("Script path %s not found" % script['path'])
            raise Exception('INVALID_SCRIPT')

    # Some G6/G7 controllers moved from the cciss subsystem to scsi
    repo_ver = repository.BaseRepository.getRepoVer(accessor)
    if repo_ver > xs_6_2:
        devices = pci.PCIDevices()
        raid_devs = devices.findByClass('01', '04')
        g6 = [x['vendor'] == '103c' and x['device'] == '323a' and
                 x['subvendor'] == '103c' and x['subdevice'].startswith('324') for x in raid_devs]
        if True in g6:
            new_device = root_device.replace('cciss', 'scsi')
            logger.info("Replacing root disk "+root_device+ " with "+new_device)
            root_device = new_device

    # Use /dev/disk/by-id/md-uuid-* rather than /dev/md*
    if repo_ver > xs_6_5 and root_device.startswith('/dev/md_'):
        uuid = get_md_uuid(root_device)
        root_device = '/dev/disk/by-id/md-uuid-%s' % uuid

    root_label, boot_label = get_fs_labels()
    if not root_label:
        logger.error("Failed to determine root label")
        raise Exception('DISK_PROBE_ERROR')

    try:
        root_partition = map_label_to_partition(root_label)
        if boot_label:
            boot_partition = map_label_to_partition(boot_label)
    except Exception as e:
        logger.logException(e)
        logger.error("Failed to map label to partition")
        raise Exception('DISK_PROBE_ERROR')
    if not root_partition:
        logger.error("Failed to determine root partition")
        raise Exception('DISK_PROBE_ERROR')
    if boot_label and not boot_partition:
        logger.error("Failed to determine boot partition")
        raise Exception('DISK_PROBE_ERROR')

    # Use /dev/disk/by-id/md-uuid-* rather than /dev/md*
    if repo_ver > xs_6_5 and root_partition.startswith('/dev/md_'):
        matched = re.search(r'[a-zA-Z](\d+)$', root_partition)
        if matched:
            partno = matched.group(1)
        else:
            logger.error("No matching partition")
            raise Exception('DISK_PROBE_ERROR')
        uuid = get_md_uuid(root_partition)
        root_partition = '/dev/disk/by-id/md-uuid-%s-part%s' % (uuid, partno)

    logger.debug("Root device: "+root_device)
    logger.debug("Root label: "+root_label)
    logger.debug("Root partition: "+root_partition)
    logger.debug("Boot label: %s" % boot_label)
    logger.debug("Boot partition: %s" % boot_partition)

    out_s = BytesIO()
    out_arc = cpiofile.CpioFile.open(fileobj=out_s, mode='w|gz')
    out_arc.hardlinks = False

    # create bootloader revert script
    config = bootloader.Bootloader.loadExisting()

    # If the existing bootloader config is already set up for an RPU, look
    # at the backup bootloader config to find the original default
    # bootloader entry.
    revert_to = config.default
    if revert_to == 'upgrade' and backup_config:
        revert_to = backup_config.default

    rc, out = cmd.runCmd(['/bin/lsblk', '-n', '-o', 'FSTYPE', root_partition], with_stdout=True)
    rootfsType = out.strip() if rc == 0 else "ext3"

    logger.info("Creating revert script")
    text = '#!/usr/bin/env python\n'
    text += '\nimport os.path\n'
    text += 'import xcp.bootloader as bootloader\n'
    text += 'import xcp.mount as mount\n'
    text += '\nrootfs = mount.TempMount("%s", "root", fstype = "%s")\n' % (root_partition, rootfsType)
    if boot_partition:
        text += 'mount.mount("%s", os.path.join(rootfs.mount_point, "boot/efi"), fstype = "vfat")\n' % boot_partition
    text += 'cfg = bootloader.Bootloader.loadExisting(rootfs.mount_point)\n'
    text += 'cfg.default = "%s"\n' % revert_to
    text += 'cfg.remove("upgrade")\n'
    text += 'cfg.commit()\n'
    if boot_partition:
        text += 'mount.umount(os.path.join(rootfs.mount_point, "boot/efi"))\n'
    text += 'rootfs.unmount()\n'

    contents = BytesIO(text.encode('utf-8'))

    f = cpiofile.CpioInfo('revert-bootloader.py')
    f.size = len(contents.getvalue())
    out_arc.addfile(f, contents)

    if url.startswith('dev://'):
        base_dir = os.path.dirname(without_scheme(url))
        path = os.path.join(base_dir, 'answerfile')
        if os.path.isfile(path):
            logger.info("Using existing answerfile")
            with open(path) as fileh:
                text = fileh.read()
    else:
        fsType = ""
        if "fs-type" in args:
            fsType = " fs-type=\"%s\"" % args.get("fs-type")

        # create answerfile
        logger.info("Creating answerfile")
        text = '<?xml version="1.0"?>\n'
        text += ' <installation mode="upgrade">\n'
        text += '  <existing-installation%s>%s</existing-installation>\n' % (fsType, root_device)
        text += '  <source type="url">%s</source>\n' % url
        platform = args.get('platform', None)
        if platform:
            text += '  <target-platform>%s</target-platform>\n' % platform
        for update in updates:
            text += '  <source type="url">%s</source>\n' % update
        text += '  <script stage="installation-start" type="url">file:///revert-bootloader.py</script>\n'
        for script in scripts:
            text += '  <script stage="%s" type="url">%s</script>\n' % (script['stage'], script['url'])

        # Add FCoE configuration if needed.
        fcoe_args = [x for x in config.menu[revert_to].kernel_args.split() if x.startswith('fcoe=')]
        for i in fcoe_args:
            m = re.match("fcoe=(.*):((no)?dcb)$", i)
            if not m:
                raise Exception('Invalid fcoe configuration')
            dcb = 'true' if m.group(2) == 'dcb' else 'false'
            text += '  <fcoe-interface hwaddr="%s" dcb="%s" />\n' % (m.group(1), dcb)

        text += ' </installation>\n'

    contents = BytesIO(text.encode('utf-8'))

    f = cpiofile.CpioInfo('answerfile')
    f.size = len(contents.getvalue())
    out_arc.addfile(f, contents)

    for script in scripts:
        if script['url'].startswith("file://"):
            logger.info("Creating %s script" % script['url'])
            f = open(script['path'], 'r')
            text = f.read()
            f.close()
            contents = BytesIO(text.encode('utf-8'))

            f = cpiofile.CpioInfo(script['name'])
            f.size = len(contents.getvalue())
            out_arc.addfile(f, contents)

    out_arc.close()

    # To avoid repacking, append the new CPIO archive to the existing
    # install.img. The kernel is capable of handling an initramfs that
    # contains multiple, optionally compressed CPIO archives
    # concatenated together, even if the compression methods are
    # different.
    f = open(installer_dir + '/install.img', 'ab')
    out_s.seek(0)
    f.write(out_s.read())
    f.close()
    out_s.close()

def resolve_bonded_iface_and_check_carrier(pif, session):
    bonds = pif.get('bond_master_of', '')
    if not bonds:
        # Return the PIF that was passed in if it is not a bond
        d = pif.get('device', '').strip()
        m = pif.get('MAC', '').strip()
        if d == '' or m == '':
            return None, None
        return d, m

    # Otherwise, determine the real interfaces behind the bond
    bond = session.xenapi.Bond.get_record(bonds[0])
    for slave in bond.get('slaves', []):
        slave_pif = session.xenapi.PIF.get_record(slave)
        device = slave_pif.get('device', '').strip()
        carrier = 0
        # Check if there is a carrier on this device
        try:
            f = open('/sys/class/net/%s/carrier' % device, 'r')
            carrier = int(f.read().strip())
            f.close()
        except Exception:
            pass
        if carrier == 1:
            m = slave_pif.get('MAC', '').strip()
            if m != '':
                return device, m

    # At this point, no interface had a carrier
    return None, None

def get_non_vlan_pif(session, device, this_host):
    """Get the non-VLAN PIF for a device on a host."""

    for pif in session.xenapi.PIF.get_all_records().values():
        if pif.get('device', '') != device:
            continue
        if pif.get('host', '') != this_host:
            continue
        if pif.get('VLAN', '-1') == '-1':
            return pif

    return None

def get_iface_config(iface):
    pif = None
    mac = None
    vlan = None

    session = XenAPI.xapi_local()
    session.xenapi.login_with_password('', '', '', 'prepare_host_upgrade.py')

    this_host = session.xenapi.session.get_this_host(session.handle)

    for net in session.xenapi.network.get_all_records().values():
        if net.get('bridge', '') == iface:
            for p in net.get('PIFs', []):
                pif = session.xenapi.PIF.get_record(p)
                if pif.get('host', '') == this_host:
                    vlan = int(pif.get('VLAN', '-1'))
                    if vlan == -1:
                        vlan = None
                        iface, mac = resolve_bonded_iface_and_check_carrier(pif, session)
                    else:
                        tmp = get_non_vlan_pif(session, pif.get('device', ''), this_host)
                        if tmp:
                            iface, mac = resolve_bonded_iface_and_check_carrier(tmp, session)
                    break

    session.xenapi.session.logout()
    return pif, iface, mac, vlan

def urlsplit(url):
    host = ''
    parts = urlparse.urlsplit(url)
    if parts.scheme in network_schemes:
        host = parts.hostname
    return (parts.scheme, host)

def set_boot_config(installer_dir, url, backup_config):
    def update_kernel_args(kernel_args, mac, config_str):
        if config.src_fmt == 'grub2':
            network_config_str = "'network_config=%s'"
        else:
            network_config_str = "network_config=%s"

        kernel_args.extend(['network_device=' + mac,
                            network_config_str % (config_str,)])

    config = bootloader.Bootloader.loadExisting()
    new_config = bootloader.Bootloader.readExtLinux(os.path.join(installer_dir, 'isolinux.cfg'))

    default = config.menu[config.default]
    new_default = new_config.menu[new_config.default]
    if 'upgrade' in config.menu_order:
        config.remove('upgrade')
        if backup_config:
            backup_config.commit(os.path.join(installer_dir, os.path.basename(backup_config.src_file)))
    else:
        if config.src_file.startswith('/boot/efi'):
            config.commit(os.path.join(installer_dir, 'efi-%s' % os.path.basename(config.src_file)))
        else:
            config.commit(os.path.join(installer_dir, os.path.basename(config.src_file)))

    xen_args = [x for x in new_default.hypervisor_args.split() if not x.startswith('com') and not x.startswith('console=')]
    xen_args.extend([x for x in default.hypervisor_args.split() if x.startswith('com') or x.startswith('console=')])
    kernel_args = [x for x in default.kernel_args.split() if x.startswith('console=') or x.startswith('xencons=') or x.startswith('device_mapper_multipath=')]
    if True in [x.startswith('com') for x in xen_args]:
        answerfile_param = 'rt_answerfile=file:///answerfile'
    else:
        answerfile_param = 'answerfile=file:///answerfile'
    kernel_args.extend(['install', answerfile_param])

    (scheme, host) = urlsplit(url)
    if scheme in network_schemes:
        # determine interface host is accessible over
        logger.debug("Repo host: "+host)
        rc, out = cmd.runCmd(['ip', 'route', 'get', socket.gethostbyname(host)],
                             with_stdout=True)
        if rc != 0:
            logger.error("Unable to resolve IP address of " + host)
            raise Exception('NETWORK_RESOLVE_ERROR')
        m = re.search(r' dev (\w+) ', out)
        if not m:
            logger.error("Unable to determine route to " + host)
            raise Exception('NETWORK_NO_ROUTE')
        iface = m.group(1)
        pif, real_iface, mac, vlan = get_iface_config(iface)
        if not pif or not real_iface or not mac:
            logger.error("Unable to determine configuration of " + iface)
            raise Exception('NETWORK_GENERIC_CONFIG_ERROR')
        logger.info("%s accessible via %s (%s)%s" %
                    (host, iface, real_iface,
                     ', vlan %s' % (vlan,) if vlan else ''))

        if pif['ip_configuration_mode'] == 'Static':
            for p in ('IP', 'gateway', 'netmask'):
                if pif.get(p, '') == '':
                    logger.error(p.capitalize()+" parameter missing for static network configuration")
                    raise Exception('NETWORK_%s_CONFIG_ERROR' % (p.upper(),))
            config_str = "static:ip=%s;netmask=%s;gateway=%s" % (pif['IP'], pif['netmask'], pif['gateway'])
            if vlan:
                config_str += ';vlan=%s' % (vlan,)
            if not re.match(r'(\d+\.){3}\d+', host) and pif.get('DNS', '') == '':
                logger.error("DNS parameter missing for static network configuration")
                raise Exception('NETWORK_DNS_CONFIG_ERROR')
            if  pif.get('DNS', '') != '':
                config_str += ";dns=" + pif['DNS']

            update_kernel_args(kernel_args, mac, config_str)
        else:
            config_str = 'dhcp'
            if vlan:
                config_str += ':vlan=%s' % (vlan,)
            update_kernel_args(kernel_args, mac, config_str)
        kernel_args.append("map_netdev=%s:d:%s" % (real_iface, mac))

    elif scheme == 'file' or \
         (scheme == 'dev' and os.path.isfile(without_scheme(url))):
        # locate major/minor of device url is on
        s = os.stat(without_scheme(url))
        major = os.major(s.st_dev)
        minor = os.minor(s.st_dev)

        # locate device name
        dev = None
        fh = open('/proc/partitions')
        fh.readline() # header line
        fh.readline() # header line
        for line in fh:
            v = line.split()
            if int(v[0]) == major and int(v[1]) == minor:
                dev = v[3]
                break
        fh.close()
        if not dev:
            logger.error("Unable to locate name for %d:%d" % (major, minor))
            raise Exception('DISK_SOURCE_NOT_FOUND')
        dev = '/dev/' + dev

        # locate mount pount
        mnt = None
        fs = None
        fh = open('/proc/mounts')
        mounted_devices = []
        for line in fh:
            v = line.split()
            if v[0] == dev:
                mnt = v[1]
                fs = v[2]
                break
            if v[0][:5] == '/dev/':
                mounted_devices.append(v)
        fh.close()
        # not found with name, try with major/minor
        if not mnt:
            for v in mounted_devices:
                try:
                    s = os.stat(v[0])
                    if not stat.S_ISBLK(s.st_mode):
                        continue
                    dev_major = os.major(s.st_rdev)
                    dev_minor = os.minor(s.st_rdev)
                    if dev_major == major and dev_minor == minor:
                        # override device, usually better than the one in /proc/partitions
                        dev = v[0]
                        mnt = v[1]
                        fs = v[2]
                        break
                except Exception as e:
                    pass
        if not mnt:
            logger.error("Unable to locate mount point for " + dev)
            raise Exception('DISK_SOURCE_NOT_FOUND')

        if scheme == 'file':
            kernel_args.append("mount=%s:%s:%s" % (dev, fs, mnt))
        else:
            kernel_args.append("make-ramdisk=%s" % dev)
            old_xen_args = xen_args
            xen_args = []
            for arg in old_xen_args:
                if arg.startswith('dom0_mem='):
                    # extract current memory size
                    m = re.match(r'dom0_mem=([^,]+),max:', arg)
                    if m:
                        mem = m.group(1)
                        mem_int = int(mem[:-1])
                        if mem.endswith('G'):
                            mem_int *= 1024

                        # calculate space needed for ramdisk
                        s = os.statvfs(without_scheme(url))
                        extra = (s.f_blocks-s.f_bfree)*s.f_frsize//(1024*1024)
                        mem_int += extra

                        # replace argument with new memory size
                        arg = "dom0_mem=%dM,max:%dM" % (mem_int, mem_int)
                xen_args.append(arg)

    try:
        # determine if boot from SAN
        f = open('/etc/firstboot.d/data/sr-multipathing.conf')
        for l in f:
            line = l.strip()
            if line.startswith('MULTIPATHING_ENABLED='):
                bfs = shell_value(line)
                if bfs == 'True':
                    kernel_args.append("device_mapper_multipath=true")
                    logger.debug("Multipathing enabled")
                break
        f.close()
    except Exception as e:
        logger.logException(e)
        logger.error("Failed to read SR multipathing config")
        raise Exception('SR_MULTIPATH_CONFIG_ERROR')

    if 'rd.iscsi.ibft=1' in default.kernel_args.split():
        logger.debug('iSCSI BFT enabled')
        kernel_args.append('use_ibft')

    root_label, _ = get_fs_labels()
    if not root_label:
        logger.error("Failed to determine root label")
        raise Exception('DISK_PROBE_ERROR')

    e = bootloader.MenuEntry(hypervisor=installer_dir + '/xen.gz', hypervisor_args=' '.join(xen_args),
                             kernel=installer_dir + '/vmlinuz', kernel_args=' '.join(kernel_args),
                             initrd=installer_dir + '/install.img', title='Rolling pool upgrade')
    e.root = root_label
    config.append('upgrade', e)
    if not config.setNextBoot('upgrade'):
        logger.info("Unable to set one-off boot, setting default entry")
        config.default = 'upgrade'
    logger.info("Writing updated bootloader config")
    config.commit()

TEST_REPO_GOOD = 0
TEST_URL_INVALID = 1
TEST_VER_INVALID = 2

def test_repo(url):
    logger.debug("Testing " + sanitize_url(url))
    repo_ver = None
    try:
        a = xcp.accessor.createAccessor(url, True)
        r = test_boot_files(a)
        if r != TEST_REPO_GOOD:
            return r
        logger.debug("Boot files ok, testing repository...")
        repo_ver = repository.BaseRepository.getRepoVer(a)
    except Exception as e:
        logger.logException(e)
        return TEST_URL_INVALID
    if not repo_ver:
        logger.error("Unable to determine repository version")
        return TEST_URL_INVALID

    # read current host version
    curr_ver = None
    try:
        i = open('/etc/xensource-inventory')
        for l in i:
            line = l.strip()
            if line.startswith('PLATFORM_VERSION='):
                curr_ver = version.Version.from_string(shell_value(line))
                break
        i.close()
    except Exception as e:
        pass

    # verify repo version
    if repo_ver and curr_ver and repo_ver >= curr_ver and curr_ver >= xs_min_version:
        logger.info("Repo version OK: " + str(repo_ver))
        return TEST_REPO_GOOD

    logger.error("Repo version ERR: " + str(repo_ver))
    return TEST_VER_INVALID

def test_update(url, main_url):
    logger.info("Testing " + sanitize_url(url))

    scheme, main_host = urlsplit(main_url)
    if scheme in network_schemes:
        # check that the host part of update URL matches the main URL
        scheme, host = urlsplit(url)
        if scheme in network_schemes:
            if host != main_host:
                logger.error("URL host mismatch: %s and %s" % (main_host, host))
                return TEST_URL_INVALID

    a = xcp.accessor.createAccessor(url, True)

    ret = TEST_URL_INVALID
    a.start()
    try:
        if a.access('update.xml'):
            logger.info("    success")
            ret = TEST_REPO_GOOD
        else:
            logger.error("    failed")
    except Exception as e:
        logger.logException(e)

    a.finish()
    return ret

def load_backup_config(installer_dir):
    try:
        if os.path.exists(os.path.join(installer_dir, 'efi-grub.cfg')):
            return bootloader.Bootloader.readGrub2(os.path.join(installer_dir, 'efi-grub.cfg'))
        elif os.path.exists(os.path.join(installer_dir, 'grub.cfg')):
            return bootloader.Bootloader.readGrub2(os.path.join(installer_dir, 'grub.cfg'))
        elif os.path.exists(os.path.join(installer_dir, 'extlinux.conf')):
            return bootloader.Bootloader.readExtLinux(os.path.join(installer_dir, 'extlinux.conf'))
    except Exception as e:
        logger.logException(e)

    return None

def prepare_host_upgrade(url, updates, scripts, args):
    installer_dir = '/boot/installer'
    backup_config = load_backup_config(installer_dir)

    # download the installer files
    shutil.rmtree(installer_dir, ignore_errors=True)
    try:
        os.mkdir(installer_dir, 0o700)
    except Exception as e:
        logger.logException(e)
        raise Exception('IOERROR')

    try:
        a = xcp.accessor.createAccessor(url, True)
        get_boot_files(a, installer_dir)

        gen_answerfile(a, installer_dir, url, updates, scripts, backup_config, args)

        # create bootloader entry
        set_boot_config(installer_dir, url, backup_config)
    except Exception as e:
        shutil.rmtree(installer_dir, ignore_errors=True)
        raise

def testSafe2Upgrade(session, args):
    return safe2upgrade(args)

def dev_from_devpath(devpath):
    """Returns the dev number of the device as a tuple."""

    devno = os.stat(devpath).st_rdev
    return os.major(devno), os.minor(devno)

def dev_from_sysfs(path):
    """Returns the dev number as a tuple from a sysfs entry."""

    with open('%s/dev' % path, 'r') as f:
        return tuple(map(int, f.read().strip().split(':')))

# The logic for this function is based on sysfs_devno_to_wholedisk in
# util-linux. See that function for reasoning.
def parentdev_from_devpath(devpath):
    """Returns the dev number of the parent device, or None if there isn't
    one."""

    try:
        devno = os.stat(devpath).st_rdev
        major = os.major(devno)
        minor = os.minor(devno)
        syspath = '/sys/dev/block/%d:%d' % (major, minor)

        partitionpath = syspath + '/partition'
        if os.path.exists(partitionpath):
            linkpath = os.path.realpath(syspath)
            parent = os.path.dirname(linkpath)
            return dev_from_sysfs(parent)
        else:
            dm_uuidpath = syspath + '/dm/uuid'
            if os.path.exists(dm_uuidpath):
                with open(dm_uuidpath, 'r') as f:
                    dm_uuid = f.read().strip()
                if re.match('part[0-9+]-', dm_uuid):
                    parent = os.listdir(syspath + '/slaves')[0]
                    return dev_from_sysfs('%s/slaves/%s' % (syspath, parent))
    except Exception as e:
        logger.logException(e)

    # If there is no parent of the parent cannot be determined...
    return None

# plugin safe upgrade test
def safe2upgrade(args):

    fh = open('/etc/xensource-inventory')
    for l in fh:
        line = l.strip()
        k, v = line.split('=', 1)
        if k == 'PRIMARY_DISK':
            primary_disk = os.path.realpath(v.strip("'"))
            break
    fh.close()

    session = XenAPI.xapi_local()
    session.xenapi.login_with_password('', '')

    this_host = session.xenapi.session.get_this_host(session.handle)

    rc, _ = cmd.runCmd(['grep', '-q', '/var/log', '/proc/mounts'], with_stdout=True)
    if rc == 0:
        return 'true'

    try:
        url = args['url']
    except KeyError:
        logger.critical("Missing argument 'url'")
        raise Exception('MISSING_URL')

    accessor = xcp.accessor.createAccessor(url, True)
    repo_ver = repository.BaseRepository.getRepoVer(accessor)
    if repo_ver >= xs_8_C:
        rc, output = cmd.runCmd(["fdisk", "-l"], with_stdout=True)
        if rc == 0:
            if "Disk label type: dos" in output:
                return 'legacy_partition_table'
            if "Dell Utility" in output:
                return 'utility_part_present'
        else:
            return 'false'

    local_sr = None
    for pbd in session.xenapi.PBD.get_all_records().values():
        if pbd.get('host', '') != this_host:
            continue
        if not pbd.get('currently_attached', False):
            continue
        devconf = pbd.get('device_config', {})
        devpath = devconf.get('device', '')
        if devpath == '':
            continue

        # From Dundee onwards, /dev/mapper paths may omit the 'p' but the old
        # path is still stored in the XAPI db. Handle this correctly.
        old_dev_mapper_match = re.match('/dev/mapper/([0-9a-fA-F]+)p([0-9]+)', devpath)
        if old_dev_mapper_match:
            new_path = '/dev/mapper/%s%s' % (old_dev_mapper_match.group(1),
                                             old_dev_mapper_match.group(2))
            if not os.path.exists(devpath) and os.path.exists(new_path):
                logger.debug('Replacing %s with %s' % (devpath, new_path))
                devpath = new_path

        devpath = os.path.realpath(devpath)
        if parentdev_from_devpath(devpath) != dev_from_devpath(primary_disk):
            continue
        local_sr = pbd['SR']
        break

    if local_sr is None:
        logger.debug("No PBD found")
        return 'true'
    else:
        logger.debug("PBD: " + local_sr)

    local_sr_size = session.xenapi.SR.get_physical_size(local_sr)
    logger.debug("PBD size: %s" % local_sr_size)
    if int(local_sr_size) < min_upgrade_lvm_part_size:
        logger.debug("PBD size smaller than minimum required")
        return 'not_enough_space'

    vdi_num = 0
    for vdi in session.xenapi.VDI.get_all_records().values():
        if vdi.get('SR', '') != local_sr:
            continue
        vdi_num += 1
    logger.debug("Number of VDIs: %d" % vdi_num)

    return 'true' if vdi_num == 0 else 'vdi_present'

# plugin url test
def testUrl(session, args):
    logger.logToSyslog(level=logging.DEBUG)

    try:
        url = args['url']
    except KeyError:
        raise Exception('MISSING_URL')

    ret = test_repo(url)

    if ret == TEST_REPO_GOOD and 'updates' in args:
        updates = args['updates'].split(';')
        for update in updates:
            ret2 = test_update(update, url)
            if ret2 != TEST_REPO_GOOD:
                ret = ret2

    if ret != TEST_REPO_GOOD:
        process_error_code(ret)

    return "true"


def getDom0DefaultMemory(session, args):
    """Returns the default amount of memory in bytes assigned to dom0.

    If a platform version is given, it returns the default amount for that
    version.
    If a URL is given, it returns the default amount for the version at the URL.
    Otherwise, it returns the default amount for the latest version."""

    logger.logToSyslog(level=logging.DEBUG)

    try:
        xc = xen.lowlevel.xc.xc()
        host_mem_kib = xc.physinfo()['total_memory']
    except Exception as e:
        logger.logException(e)
        raise Exception('UNKNOWN_ERROR')

    if 'version' in args:
        v = version.Version.from_string(args['version'])
        return str(dom0.default_memory_for_version(host_mem_kib, v) * 1024)
    elif 'url' in args:
        logger.debug("Getting version from " + sanitize_url(args['url']))
        a = None
        try:
            a = xcp.accessor.createAccessor(args['url'], True)
            v = repository.BaseRepository.getRepoVer(a)
            return str(dom0.default_memory_for_version(host_mem_kib, v) * 1024)
        except Exception as e:
            logger.logException(e)
            if a is not None and a.lastError:
                logger.error(a.lastError)
                process_error_code(a.lastError)
            else:
                raise Exception('UNKNOWN_ERROR')
    else:
        return str(dom0.default_memory(host_mem_kib) * 1024)


def getVersion(session, args):
    logger.logToSyslog(level=logging.DEBUG)

    try:
        url = args['url']
    except KeyError:
        raise Exception('MISSING_URL')

    logger.debug("Getting version from " + sanitize_url(url))
    a = None
    try:
        a = xcp.accessor.createAccessor(url, True)
        platform_ver = repository.BaseRepository.getRepoVer(a)
        product_ver = repository.BaseRepository.getProductVersion(a)
        return json.dumps({'platform-version': str(platform_ver) if platform_ver else None,
                           'product-version': str(product_ver) if product_ver else None})
    except Exception as e:
        logger.logException(e)
        if a is not None and a.lastError:
            logger.error(a.lastError)
            process_error_code(a.lastError)
        else:
            raise Exception('UNKNOWN_ERROR')

def getEUA(session, args):
    logger.logToSyslog(level=logging.DEBUG)

    eua_paths = ['EUA', 'EULA']
    eua_content = None
    try:
        url = args['url']
    except KeyError:
        raise Exception('MISSING_URL')

    logger.debug("Getting EUA from " + sanitize_url(url))
    a = None
    try:
        a = xcp.accessor.createAccessor(url, True)
        a.start()
        for eua_path in eua_paths:
            if a.access(eua_path):
                fp = a.openAddress(eua_path)
                eua_content = fp.read().decode('utf-8')
                fp.close()
                break
            else:
                logger.error("Can not access EUA file")
    except Exception as e:
        logger.logException(e)
        if a is not None and a.lastError:
            logger.error(a.lastError)
            process_error_code(a.lastError)
        else:
            raise Exception('UNKNOWN_ERROR')

    a.finish()
    return json.dumps({'eua': str(eua_content) if eua_content else None})

# plugin entry point
def main(session, args):
    logger.logToSyslog(level=logging.DEBUG)

    try:
        url = args['url']
    except KeyError:
        logger.critical("Missing argument 'url'")
        raise Exception('MISSING_URL')

    updates = []
    if 'updates' in args:
        updates = args['updates'].split(";")
    scripts = []
    if 'script' in args:
        scripts_args = args['script'].split(":")
        if len(scripts_args) < 2 or len(scripts_args) % 2 != 0:
            logger.critical("Value of argument 'script' must be in pairs stage1:path1[:stageN:pathN]")
            raise Exception("INVALID_SCRIPT")

        for script_args in zip(scripts_args[0::2], scripts_args[1::2]):
            if not script_args[0] or not script_args[1]:
                logger.critical("Missing stage or path in 'script' arguments (%s)" % (script_args,))
                raise Exception("INVALID_SCRIPT")
            script_stage, script_path = script_args
            script_name = os.path.basename(script_path)
            if "://" in script_path:
                script_url = script_path
            else:
                script_url = "file:///" + script_name
            script = {'stage': script_stage, 'path': script_path, 'name': script_name, 'url': script_url}
            scripts.append(script)

    logger.info("Verifying repo...")
    rc = test_repo(url)
    if rc != TEST_REPO_GOOD:
        logger.error("%s is not a valid repo" % sanitize_url(url))
        process_error_code(rc)

    logger.info("Repo ok, preparing for upgrade")
    try:
        prepare_host_upgrade(url, updates, scripts, args)
    except Exception as e:
        # Log the traceback so it is obvious where it came from.
        logger.error("There was an error in preparing the host for upgrade.")
        logger.logException(e)
        raise

    if safe2upgrade(args) == 'true':
        fh = open('/var/preserve/safe2upgrade', 'w')
        fh.close()
    else:
        if os.path.isfile('/var/preserve/safe2upgrade'):
            os.remove('/var/preserve/safe2upgrade')

    logger.info("Preparation succeeded, ready for upgrade.")
    return "true"


if __name__ == '__main__':
    XenAPIPlugin.dispatch({"getDom0DefaultMemory": getDom0DefaultMemory,
                           "getVersion": getVersion,
                           "getEUA": getEUA,
                           "main": main,
                           "testUrl": testUrl,
                           "testSafe2Upgrade": testSafe2Upgrade})
